from __future__ import absolute_import
import venusian
from pyramid.decorator import reify
from pyramid.traversal import find_root
from types import MethodType
from .interfaces import (
    CALCULATED_PROPERTIES,
    CONNECTION,
)


def includeme(config):
    config.registry[CALCULATED_PROPERTIES] = CalculatedProperties()
    config.add_directive('add_calculated_property', add_calculated_property)


class ItemNamespace(object):
    def __init__(self, context, request, defined=None, ns=None):
        self.context = context
        self.request = request
        self._defined = defined or {}
        if ns:
            self.__dict__.update(ns)
        self._results = {}

    @reify
    def _properties(self):
        return self.context.__json__(self.request)

    @reify
    def root(self):
        return find_root(self.context)

    @reify
    def registry(self):
        return self.request.registry

    def __getattr__(self, name):
        context = self.context
        request = self.request
        conn = self.registry[CONNECTION]
        if name in self._defined:
            value = self._defined[name](self)
            setattr(self, name, value)
            return value
        if name in self._properties:
            value = self._properties[name]
            if name in context.type_info.schema_links:
                if isinstance(value, list):
                    value = [
                        request.resource_path(conn.get_by_uuid(v))
                        for v in value
                    ]
                else:
                    value = request.resource_path(conn.get_by_uuid(value))
            setattr(self, name, value)
            return value
        if name in context.rev:
            value = context.get_rev_links(request, name)
            value = [
                request.resource_path(conn.get_by_uuid(v))
                for v in value
            ]
            setattr(self, name, value)
            return value
        raise AttributeError(name)

    def __call__(self, fn):
        try:
            return self._results[fn]
        except KeyError:
            pass

        if isinstance(fn, str):
            result = self._results[fn] = getattr(self, fn, None)
            return result

        start = 1 if isinstance(fn, MethodType) else 0
        # Not using inspect.getargspec as it is slow
        args = fn.__code__.co_varnames[start:fn.__code__.co_argcount]
        kw = {}
        for name in args:
            try:
                kw[name] = getattr(self, name)
            except AttributeError:
                pass

        result = self._results[fn] = fn(**kw)
        return result


class CalculatedProperties(object):
    def __init__(self):
        self.category_cls_props = {}

    def register_prop(self, fn, name, context, condition=None, schema=None,
                      attr=None, define=False, category='object'):
        prop = CalculatedProperty(fn, name, attr, condition, schema, define)
        cls_props = self.category_cls_props.setdefault(category, {})
        cls_props.setdefault(context, {})[name] = prop

    def props_for(self, context, category='object'):
        if isinstance(context, type):
            cls = context
        else:
            cls = type(context)
        props = {}
        cls_props = self.category_cls_props.get(category, {})
        for base in reversed(cls.mro()):
            props.update(cls_props.get(base, {}))
        return props


class CalculatedProperty(object):
    condition_args = None

    def __init__(self, fn, name, attr=None, condition=None, schema=None, define=False):
        self.fn = fn
        self.attr = attr
        self.name = name
        self.condition = condition
        self.define = define

        if schema is not None:
            if 'default' in schema:
                raise ValueError('schema may not specify default for calculated property')
            schema = schema.copy()
            schema['calculatedProperty'] = True
        self.schema = schema

    def __call__(self, namespace):
        if self.condition is not None:
            if not namespace(self.condition):
                return None
        if self.attr:
            fn = getattr(namespace.context, self.attr)
        else:
            fn = self.fn
        return namespace(fn)


# Imperative configuration
def add_calculated_property(config, fn, name, context, condition=None, schema=None,
                            attr=None, define=False, category='object'):
    calculated_properties = config.registry[CALCULATED_PROPERTIES]
    config.action(
        ('calculated_property', context, category, name),
        calculated_properties.register_prop,
        (fn, name, context, condition, schema, attr, define, category),
    )


# Declarative configuration
def calculated_property(**settings):
    """ Register a calculated property
    """

    def decorate(wrapped):
        def callback(scanner, factory_name, factory):
            if settings.get('context') is None:
                settings['context'] = factory
            if settings.get('name') is None:
                settings['name'] = factory_name
            scanner.config.add_calculated_property(wrapped, **settings)

        info = venusian.attach(wrapped, callback, category='object')

        if info.scope == 'class':
            # if the decorator was attached to a method in a class, or
            # otherwise executed at class scope, we need to set an
            # 'attr' into the settings if one isn't already in there
            if settings.get('attr') is None:
                settings['attr'] = wrapped.__name__
            if settings.get('name') is None:
                settings['name'] = wrapped.__name__

        elif settings.get('context') is None:
            raise TypeError('must supply context type for function')

        return wrapped

    return decorate


def calculate_properties(context, request, ns=None, category='object'):
    calculated_properties = request.registry[CALCULATED_PROPERTIES]
    props = calculated_properties.props_for(context, category)
    defined = {name: prop for name, prop in props.items() if prop.define}
    if isinstance(context, type):
        context = None
    namespace = ItemNamespace(context, request, defined, ns)
    return {
        name: value
        for name, value in (
            (name, prop(namespace))
            for name, prop in props.items()
        ) if value is not None
    }



import logging
import pytest
import sqlalchemy
import transaction as transaction_management
import webtest
import webtest.http
import zope.sqlalchemy

from contextlib import contextmanager
from transaction.interfaces import ISynchronizer
from urllib.parse import quote
from zope.interface import implementer

from ..app import configure_engine
from ..storage import Base
from .elasticsearch_fixture import server_process as elasticsearch_server_process
from .postgresql_fixture import initdb, server_process as postgres_server_process


def pytest_configure():
    logging.basicConfig(format='')
    logging.getLogger('sqlalchemy.engine').setLevel(logging.WARNING)

    class Shorten(logging.Filter):
        max_len = 500

        def filter(self, record):
            if record.msg == '%r':
                record.msg = record.msg % record.args
                record.args = ()
            if len(record.msg) > self.max_len:
                record.msg = record.msg[:self.max_len] + '...'
            return True

    logging.getLogger('sqlalchemy.engine.base.Engine').addFilter(Shorten())


@pytest.mark.fixture_cost(10)
@pytest.yield_fixture(scope='session')
def engine_url(request):
    # Ideally this would use a different database on the same postgres server
    tmpdir = request.config._tmpdirhandler.mktemp('postgresql-engine', numbered=True)
    tmpdir = str(tmpdir)
    initdb(tmpdir)
    process = postgres_server_process(tmpdir)

    yield 'postgresql://postgres@:5432/postgres?host=%s' % quote(tmpdir)

    if process.poll() is None:
        process.terminate()
        process.wait()


@pytest.mark.fixture_cost(10)
@pytest.yield_fixture(scope='session')
def postgresql_server(request):
    tmpdir = request.config._tmpdirhandler.mktemp('postgresql', numbered=True)
    tmpdir = str(tmpdir)
    initdb(tmpdir)
    process = postgres_server_process(tmpdir)

    yield 'postgresql://postgres@:5432/postgres?host=%s' % quote(tmpdir)

    if process.poll() is None:
        process.terminate()


@pytest.fixture(scope='session')
def elasticsearch_host_port():
    return webtest.http.get_free_port()


@pytest.mark.fixture_cost(10)
@pytest.yield_fixture(scope='session')
def elasticsearch_server(request, elasticsearch_host_port, remote_es):
    if not remote_es:
        # spawn a new one
        host, port = elasticsearch_host_port
        tmpdir = request.config._tmpdirhandler.mktemp('elasticsearch', numbered=True)
        tmpdir = str(tmpdir)
        process = elasticsearch_server_process(str(tmpdir), host=host, port=port)

        yield 'http://%s:%d' % (host, port)

        if process.poll() is None:
            process.terminate()
            process.wait()
    else:
        yield remote_es


# http://docs.sqlalchemy.org/en/rel_0_8/orm/session.html#joining-a-session-into-an-external-transaction
# By binding the SQLAlchemy Session to an external transaction multiple testapp
# requests can be rolled back at the end of the test.

@pytest.yield_fixture(scope='session')
def conn(engine_url):

    engine_settings = {
        'sqlalchemy.url': engine_url,
    }

    engine = configure_engine(engine_settings)
    conn = engine.connect()
    tx = conn.begin()
    try:
        Base.metadata.create_all(bind=conn)
        yield conn
    finally:
        tx.rollback()
        conn.close()
        engine.dispose()


@pytest.fixture(scope='session')
def _DBSession(conn):
    # ``server`` thread must be in same scope
    DBSession = sqlalchemy.orm.scoped_session(sqlalchemy.orm.sessionmaker(bind=conn), scopefunc=lambda: 0)
    zope.sqlalchemy.register(DBSession)
    return DBSession


@pytest.fixture(scope='session')
def DBSession(_DBSession, zsa_savepoints, check_constraints):
    return _DBSession


@pytest.yield_fixture
def external_tx(request, conn):
    # print('BEGIN external_tx')
    tx = conn.begin_nested()
    yield tx
    tx.rollback()
    # # The database should be empty unless a data fixture was loaded
    # for table in Base.metadata.sorted_tables:
    #     assert conn.execute(table.count()).scalar() == 0


@pytest.fixture
def transaction(request, external_tx, zsa_savepoints, check_constraints):
    transaction_management.begin()
    request.addfinalizer(transaction_management.abort)
    return transaction_management


@pytest.yield_fixture(scope='session')
def zsa_savepoints(conn):
    """ Place a savepoint at the start of the zope transaction

    This means failed requests rollback to the db state when they began rather
    than that at the start of the test.
    """

    @implementer(ISynchronizer)
    class Savepoints(object):
        def __init__(self, conn):
            self.conn = conn
            self.sp = None
            self.state = None

        def beforeCompletion(self, transaction):
            pass

        def afterCompletion(self, transaction):
            # txn be aborted a second time in manager.begin()
            if self.sp is None:
                return
            if self.state == 'commit':
                self.state = 'completion'
                self.sp.commit()
            else:
                self.state = 'abort'
                self.sp.rollback()
            self.sp = None
            self.state = 'done'

        def newTransaction(self, transaction):
            self.state = 'new'
            self.sp = self.conn.begin_nested()
            self.state = 'begun'
            transaction.addBeforeCommitHook(self._registerCommit)

        def _registerCommit(self):
            self.state = 'commit'

    zsa_savepoints = Savepoints(conn)

    transaction_management.manager.registerSynch(zsa_savepoints)

    yield zsa_savepoints
    transaction_management.manager.unregisterSynch(zsa_savepoints)


@pytest.fixture
def session(transaction, DBSession):
    """ Returns a setup session

    Depends on transaction as storage relies on some interaction there.
    """
    return DBSession()


@pytest.yield_fixture(scope='session')
def check_constraints(conn, _DBSession):
    '''Check deffered constraints on zope transaction commit.

    Deferred foreign key constraints are only checked at the outer transaction
    boundary, not at a savepoint. With the Pyramid transaction bound to a
    subtransaction check them manually.
    '''

    @implementer(ISynchronizer)
    class CheckConstraints(object):
        def __init__(self, conn):
            self.conn = conn
            self.state = None

        def beforeCompletion(self, transaction):
            pass

        def afterCompletion(self, transaction):
            pass

        def newTransaction(self, transaction):

            @transaction.addBeforeCommitHook
            def set_constraints():
                self.state = 'checking'
                session = _DBSession()
                session.flush()
                sp = self.conn.begin_nested()
                try:
                    self.conn.execute('SET CONSTRAINTS ALL IMMEDIATE')
                except:
                    sp.rollback()
                    raise
                else:
                    self.conn.execute('SET CONSTRAINTS ALL DEFERRED')
                finally:
                    sp.commit()
                    self.state = None

    check_constraints = CheckConstraints(conn)

    transaction_management.manager.registerSynch(check_constraints)

    yield check_constraints

    transaction_management.manager.unregisterSynch(check_constraints)


@pytest.yield_fixture
def execute_counter(conn, zsa_savepoints, check_constraints):
    """ Count calls to execute
    """

    class Counter(object):
        def __init__(self):
            self.reset()
            self.conn = conn

        def reset(self):
            self.count = 0

        @contextmanager
        def expect(self, count):
            start = self.count
            yield
            difference = self.count - start
            assert difference == count

    counter = Counter()

    @sqlalchemy.event.listens_for(conn, 'after_cursor_execute')
    def after_cursor_execute(conn, cursor, statement, parameters, context, executemany):
        # Ignore the testing savepoints
        if zsa_savepoints.state != 'begun' or check_constraints.state == 'checking':
            return
        counter.count += 1

    yield counter

    sqlalchemy.event.remove(conn, 'after_cursor_execute', after_cursor_execute)


@pytest.yield_fixture
def no_deps(conn, DBSession):

    session = DBSession()

    @sqlalchemy.event.listens_for(session, 'after_flush')
    def check_dependencies(session, flush_context):
        assert not flush_context.cycles

    @sqlalchemy.event.listens_for(conn, "before_execute", retval=True)
    def before_execute(conn, clauseelement, multiparams, params):
        return clauseelement, multiparams, params

    yield

    sqlalchemy.event.remove(session, 'before_flush', check_dependencies)

@pytest.fixture(scope='session')
def wsgi_server_host_port():
    return webtest.http.get_free_port()


@pytest.fixture(scope='session')
def wsgi_server_app(app):
    return app


@pytest.mark.fixture_cost(100)
@pytest.yield_fixture(scope='session')
def wsgi_server(request, wsgi_server_app, wsgi_server_host_port):
    host, port = wsgi_server_host_port

    server = webtest.http.StopableWSGIServer.create(
        wsgi_server_app,
        host=host,
        port=port,
        threads=1,
        channel_timeout=60,
        cleanup_interval=10,
        expose_tracebacks=True,
    )
    assert server.wait()

    yield 'http://%s:%s' % wsgi_server_host_port

    server.shutdown()



import os
import pytest

from base64 import b64encode
from jsonschema_serialize_fork import Draft4Validator
from pyramid.compat import ascii_native_
from .. import TYPES
from .toolfixtures import registry, root


TYPE_NAMES = ['TestingPostPutPatchSno', 'TestingDownload']


def get_parameterized_names():
    """ Get all item types from schema names """
    return [name.split('.')[0] for name in os.listdir(os.getcwd() + '/snovault/test_schemas')]

PARAMETERIZED_NAMES = get_parameterized_names()


@pytest.mark.parametrize('item_type', PARAMETERIZED_NAMES)
def test_collections(testapp, item_type):
    """ Get all item types, check they are in the response """
    res = testapp.get('/' + item_type).follow(status=200)
    assert item_type.encode('utf-8') in res.body


@pytest.mark.parametrize('item_type', PARAMETERIZED_NAMES)
def test_json(testapp, item_type):
    """ Check that when we get proper item types """
    res = testapp.get('/' + item_type).follow(status=200)
    assert (item_type + 'Collection') in res.json['@type']


def test_json_basic_auth(anontestapp):
    url = '/'
    value = "Authorization: Basic %s" % ascii_native_(b64encode(b'nobody:pass'))
    res = anontestapp.get(url, headers={'Authorization': value}, status=401)
    assert res.content_type == 'application/json'


def test_home_json(testapp):
    res = testapp.get('/', status=200)
    assert res.json['@type']


def test_collection_post_bad_json(testapp):
    item = {'foo': 'bar'}
    res = testapp.post_json('/embedding-tests', item, status=422)
    assert res.json['status'] == 'error'


def test_collection_post_malformed_json(testapp):
    item = '{'
    headers = {'Content-Type': 'application/json'}
    res = testapp.post('/embedding-tests', item, status=400, headers=headers)
    assert res.json['detail'].startswith('Expecting')


def test_collection_post_missing_content_type(testapp):
    item = '{}'
    testapp.post('/embedding-tests', item, status=415)


def test_collection_post_bad(anontestapp):
    value = "Authorization: Basic %s" % ascii_native_(b64encode(b'nobody:pass'))
    anontestapp.post_json('/embedding-tests', {}, headers={'Authorization': value}, status=401)


@pytest.mark.slow
def test_collection_limit(testapp):
    """ Post 3 EmbeddingTests, check that limit=all, limit=2 works """
    obj1 = {
        'title': "Testing1",
        'description': "This is testig object 1",
    }
    obj2 = {
        'title': "Testing2",
        'description': "This is testig object 2",
    }
    obj3 = {
        'title': "Testing3",
        'description': "This is testig object 3",
    }
    testapp.post_json('/embedding-tests', obj1, status=201)
    testapp.post_json('/embedding-tests', obj2, status=201)
    testapp.post_json('/embedding-tests', obj3, status=201)
    res_all = testapp.get('/embedding-tests/?limit=all', status=200)
    res_2 = testapp.get('/embedding-tests/?limit=2', status=200)
    assert len(res_all.json['@graph']) == 3
    assert len(res_2.json['@graph']) == 2

def test_collection_put(testapp, execute_counter):
    """ Insert and udpate an item into a collection, verify it worked """
    initial = {
        'title': "Testing",
        'type': "object", # include a non-required field
        'description': "This is the initial insert",
    }
    item_url = testapp.post_json('/embedding-tests', initial).location

    with execute_counter.expect(1):
        item = testapp.get(item_url).json

    for key in initial:
        assert item[key] == initial[key]

    update = {
        'title': "New Testing",
        'type': "object",
        'description': "This is the updated insert",
    }
    testapp.put_json(item_url, update, status=200)

    res = testapp.get('/' + item['uuid']).follow().json

    for key in update:
        assert res[key] == update[key]


def test_invalid_collection_put(testapp):
    """ Tests that inserting various invalid items will appropriately fail """
    missing_required = {
        'title': "Testing",
        'type': "object"
    }
    testapp.post_json('/embedding-tests', missing_required, status=422)

    nonexistent_field = {
        'title': "Testing",
        'type': "string",
        'descriptionn': "This is a descriptionn", # typo
    }
    testapp.post_json('/embedding-tests', nonexistent_field, status=422)

    valid = {
        'title': "Testing",
        'type': "object",
        'description': "This is a valid object",
    }
    invalid_update = {
        'descriptionn': "This is an invalid update",
    }
    item_url = testapp.post_json('/embedding-tests', valid, status=201).location
    testapp.put_json(item_url, invalid_update, status=422)


def test_page_toplevel(anontestapp):
    res = anontestapp.get('/embedding-tests/', status=200)
    assert res.json['@id'] == '/embedding-tests/'


def test_jsonld_context(testapp):
    res = testapp.get('/terms/', status=200)
    assert res.json


@pytest.mark.slow
@pytest.mark.parametrize('item_type', PARAMETERIZED_NAMES)
def test_index_data_workbook(testapp, indexer_testapp, item_type):
    res = testapp.get('/%s?limit=all' % item_type).follow(status=200)
    for item in res.json['@graph']:
        indexer_testapp.get(item['@id'] + '@@index-data', status=200)


@pytest.mark.xfail
def test_abstract_collection(testapp, experiment):
    testapp.get('/Dataset/{accession}'.format(**experiment))
    testapp.get('/datasets/{accession}'.format(**experiment))


def test_home(testapp):
    testapp.get('/', status=200)


@pytest.mark.parametrize('item_type', TYPE_NAMES)
def test_profiles(testapp, item_type):
    res = testapp.get('/profiles/%s.json' % item_type).maybe_follow(status=200)
    errors = Draft4Validator.check_schema(res.json)
    assert not errors
    # added from ..schema_views._annotated_schema
    assert 'rdfs:seeAlso' in res.json
    assert 'rdfs:subClassOf' in res.json
    assert 'children' in res.json
    assert res.json['isAbstract'] is False


@pytest.mark.parametrize('item_type', ['AbstractItemTest'])
def test_profiles_abstract(testapp, item_type):
    res = testapp.get('/profiles/%s.json' % item_type).maybe_follow(status=200)
    errors = Draft4Validator.check_schema(res.json)
    assert not errors
    # added from ..schema_views._annotated_schema
    assert 'rdfs:seeAlso' in res.json
    # Item/item does not have subClass
    if item_type.lower() == 'item':
        assert 'rdfs:subClassOf' not in res.json
    else:
        assert 'rdfs:subClassOf' in res.json
    # abstract types wil have children
    assert len(res.json['children']) >= 1
    assert res.json['isAbstract'] is True


def test_profiles_all(testapp, registry):
    res = testapp.get('/profiles/').maybe_follow(status=200)
    # make sure all types are present, including abstract types
    for ti in registry[TYPES].by_item_type.values():
        assert ti.name in res.json
    for ti in registry[TYPES].by_abstract_type.values():
        assert ti.name in res.json



from .. import Item, collection
from ..upgrader import upgrade_step, upgrade_finalizer


def includeme(config):
    config.scan(__name__)
    config.add_upgrade('testing_upgrader', '3')


@collection('testing-upgrader')
class TestingUpgrader(Item):
    item_type = 'testing_upgrader'


@upgrade_step('testing_upgrader', '', '2')
def step1(value, system):
    value['step1'] = True
    return value


@upgrade_step('testing_upgrader', '2', '3')
def step2(value, system):
    value['step2'] = True
    return value


@upgrade_finalizer('testing_upgrader')
def finalizer(value, system, version):
    value['schema_version'] = version
    return value



from pyramid.httpexceptions import (
    HTTPNotModified,
    HTTPPreconditionFailed,
)


def etag_app_version(view_callable):
    def wrapped(context, request):
        etag = request.registry.settings['snovault.app_version']
        if etag in request.if_none_match:
            raise HTTPNotModified()
        result = view_callable(context, request)
        request.response.etag = etag
        return result

    return wrapped


def etag_app_version_effective_principals(view_callable):
    def wrapped(context, request):
        app_version = request.registry.settings['snovault.app_version']
        etag = app_version + ' ' + ' '.join(sorted(request.effective_principals))
        if etag in request.if_none_match:
            raise HTTPNotModified()
        result = view_callable(context, request)
        request.response.etag = etag
        cache_control = request.response.cache_control
        cache_control.private = True
        cache_control.max_age = 0
        cache_control.must_revalidate = True
        return result

    return wrapped



import atexit
import logging
import structlog
import time
import transaction
import signal
import zope.sqlalchemy

from contextlib import contextmanager
from functools import partial
from multiprocessing import (
    get_context,
    cpu_count
)
from multiprocessing.pool import Pool
from pyramid.request import apply_request_extensions
from pyramid.threadlocal import (
    get_current_request,
    manager
)
from sqlalchemy import orm

from .. import DBSESSION, set_logging
from ..app import configure_engine
from .indexer import (
    INDEXER,
    Indexer,
)
from .interfaces import (
    APP_FACTORY,
    INDEXER_QUEUE
)
from ..storage import register_storage, RDBStorage


log = structlog.getLogger(__name__)

def includeme(config):
    if config.registry.settings.get('indexer_worker'):
        return
    config.registry[INDEXER] = MPIndexer(config.registry)


### Running in subprocess

app = None

def initializer(app_factory, settings):
    """
    Need to initialize the app for the subprocess.
    As part of this, configue a new database engine and set logging
    """
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    # set up global variables to use throughout subprocess
    global db_engine
    db_engine = None
    atexit.register(clear_manager_and_dispose_engine)

    global app
    app = app_factory(settings, indexer_worker=True, create_tables=False)
    db_engine = configure_engine(settings)

    # Use `es_server=app.registry.settings.get('elasticsearch.server')` when ES logging is working
    set_logging(in_prod=app.registry.settings.get('production'))
    global log
    log = structlog.get_logger(__name__)


@contextmanager
def threadlocal_manager():
    """
    Set registry and request attributes using the global app within the
    subprocess
    """

    # clear threadlocal manager to get a clean stack
    manager.clear()

    registry = app.registry
    request = app.request_factory.blank('/_indexing_pool')
    request.registry = registry
    request.datastore = 'database'
    apply_request_extensions(request)
    request.invoke_subrequest = app.invoke_subrequest
    request.root = app.root_factory(request)
    request._stats = getattr(request, "_stats", {})

    # configure a sqlalchemy session and set isolation level
    DBSession = orm.scoped_session(orm.sessionmaker(bind=db_engine))
    request.registry[DBSESSION] = DBSession
    # configue RDBStorage. Overide write storage to use new DBSession
    register_storage(request.registry, write_override=RDBStorage(DBSession))
    zope.sqlalchemy.register(DBSession)
    connection = DBSession().connection()
    connection.execute('SET TRANSACTION ISOLATION LEVEL REPEATABLE READ READ ONLY')

    # add the newly created request to the pyramid threadlocal manager
    manager.push({'request': request, 'registry': registry})
    yield
    # remove the session when leaving contextmanager
    request.registry[DBSESSION].remove()


def clear_manager_and_dispose_engine(signum=None, frame=None):
    manager.clear()
    # manually dispose of db engine for garbage collection
    if db_engine is not None:
        db_engine.dispose()


### These helper functions are needed for multiprocessing

def sync_update_helper(uuid):
    """
    Used with synchronous indexing. Counter is controlled at a higher level
    (MPIndexer.update_objects)
    """
    with threadlocal_manager():
        request = get_current_request()
        indexer = request.registry[INDEXER]
        return indexer.update_object(request, uuid)


def queue_update_helper():
    """
    Used with the queue. Keeps a local counter and errors, which are returned
    to the callback function and synchronized with overall values.
    `local_deferred` is True when the indexer hits an sid exception and must
    defer the indexing; it should stay as the third returned value in the
    tuple and is used in overall MPIndexer.update_objects function
    """
    with threadlocal_manager():
        local_counter = [0]
        request = get_current_request()
        indexer = request.registry[INDEXER]
        local_errors, local_deferred = indexer.update_objects_queue(request, local_counter)
        return (local_errors, local_counter, local_deferred)


def queue_error_callback(cb_args, counter, errors):
    """
    Update the counter and errors with the result of the given callback arguments
    """
    local_errors, local_counter, _ = cb_args
    if counter:
        counter[0] = local_counter[0] + counter[0]
    errors.extend(local_errors)


### Running in main process

class MPIndexer(Indexer):
    def __init__(self, registry):
        super(MPIndexer, self).__init__(registry)
        self.chunksize = int(registry.settings.get('indexer.chunk_size', 1024))
        # use 2 fewer processes than cpu count, with a minimum of 1
        num_cpu = cpu_count()
        self.processes = num_cpu - 2 if num_cpu - 2 > 1 else 1
        self.initargs = (registry[APP_FACTORY], registry.settings,)

    def init_pool(self):
        """
        Initialize multiprocessing pool.
        Use `maxtasksperchild=1`, which causes the worker to be recycled after
        finishing one call to `queue_update_helper`.
        It seems like this should not be needed due to requests being
        created by `threadlocal_manager`, but the transaction scope is not
        correctly reset on each call to `update_objects` without it.

        TODO: figure out how to remove `maxtasksperchild=1` w.r.t. pyramid_tm
              so that transaction scope is correctly handled and we can skip
              work done by `initializer` for each `queue_update_helper` call
        """
        return Pool(
            processes=self.processes,
            initializer=initializer,
            initargs=self.initargs,
            maxtasksperchild=1,  # see rationale in function documentation above.
            context=get_context('spawn'),
        )

    def update_objects(self, request, counter):
        """
        Initializes a multiprocessing pool with args given in __init__ and
        indexes in one of two mode: synchronous or queued.
        If a list of uuids is passed in the request, sync indexing will occur,
        breaking the list up among all available workers in the pool.
        Otherwise, all available workers will asynchronously pull uuids of the
        queue for indexing (see indexer.py).
        Note that counter is a length 1 array (so it can be passed by reference)
        Close the pool at the end of the function and return list of errors.
        """
        pool = self.init_pool()
        sync_uuids = request.json.get('uuids', None)
        workers = self.processes
        # ensure workers != 0
        workers = 1 if workers == 0 else workers
        errors = []

        # use sync_uuids with imap_unordered for synchronous indexing OR
        # apply_async for asynchronous indexing
        if sync_uuids:
            # determine how many uuids should be used for each process
            chunkiness = int((len(sync_uuids) - 1) / workers) + 1
            if chunkiness > self.chunksize:
                chunkiness = self.chunksize
            # imap_unordered to hopefully shuffle item types and come up with
            # a more or less equal workload for each process
            for error in pool.imap_unordered(sync_update_helper, sync_uuids, chunkiness):
                if error is not None:
                    errors.append(error)
                else:
                    counter[0] += 1  # don't increment counter on an error
                if counter[0] % 10 == 0:
                    log.info('Indexing %d (sync)', counter[0])
        else:
            # use partial here so the callback can use counter and errors
            callback_w_errors = partial(queue_error_callback, counter=counter, errors=errors)
            # hold AsyncResult objects returned by apply_async
            async_results = []
            # last_count used to track if there is "more" work to do
            last_count = 0

            # create the initial workers (same as number of processes in pool)
            for i in range(workers):
                res = pool.apply_async(queue_update_helper,
                                       callback=callback_w_errors)
                async_results.append(res)

            # check worker statuses
            # add more jobs if any are finished and indexing is ongoing
            while True:
                results_to_add = []
                idxs_to_rm = []
                for idx, res in enumerate(async_results):
                    if res.ready():
                        # res_vals are returned from one run of `queue_update_helper`
                        # in form: (errors <list>, counter <list>, deferred <bool>)
                        res_vals = res.get()
                        idxs_to_rm.append(idx)
                        # add jobs if overall counter has increased OR process is deferred
                        if (counter[0] > last_count) or res_vals[2] is True:
                            last_count = counter[0]
                            res = pool.apply_async(queue_update_helper,
                                                   callback=callback_w_errors)
                            results_to_add.append(res)

                for idx in sorted(idxs_to_rm, reverse=True):
                    del async_results[idx]
                async_results.extend(results_to_add)

                if len(async_results) == 0:
                    break
                time.sleep(0.5)

        pool.close()
        pool.join()
        return errors



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['snovault', 'snovault.commands', 'snovault.elasticsearch', 'snovault.tests']

package_data = \
{'': ['*'], 'snovault': ['test_schemas/*']}

install_requires = \
['MarkupSafe==0.23',
 'PasteDeploy==1.5.2',
 'Pillow==3.3.0',
 'PyBrowserID==0.11.0',
 'PyYAML==3.12',
 'SPARQLWrapper==1.7.6',
 'SQLAlchemy==1.2.16',
 'WSGIProxy2==0.4.2',
 'WebOb==1.8.5',
 'WebTest==2.0.21',
 'aws-requests-auth==0.4.1',
 'awscli==1.15.42',
 'bcrypt==3.1.0',
 'beautifulsoup4==4.4.1',
 'boto3==1.7.42',
 'botocore==1.10.42',
 'certifi==2018.11.29',
 'cffi==1.7.0',
 'chardet==3.0.4',
 'colorama==0.3.7',
 'coverage==4.1',
 'cryptacular==1.4.1',
 'dcicutils==0.8.5',
 'docutils==0.12',
 'elasticsearch-dsl==5.3.0',
 'elasticsearch==5.5.3',
 'flaky==3.6.1',
 'future==0.15.2',
 'futures==3.1.1',
 'html5lib==1.0b8',
 'humanfriendly==1.44.7',
 'hupper==1.5',
 'idna==2.7',
 'isodate==0.5.4',
 'jmespath==0.9.0',
 'jsonschema-serialize-fork==2.1.1',
 'keepalive==0.5',
 'loremipsum==1.0.5',
 'netaddr==0.7.18',
 'passlib==1.6.5',
 'pbkdf2==1.3',
 'plaster-pastedeploy==0.6',
 'plaster==1.0',
 'psutil==4.3.0',
 'psycopg2==2.7.3',
 'py==1.4.31',
 'pyasn1==0.1.9',
 'pycparser==2.14',
 'pyparsing==2.2.0',
 'pyramid-localroles==0.1',
 'pyramid-multiauth==0.8.0',
 'pyramid-retry==1.0',
 'pyramid-tm==2.2.1',
 'pyramid-translogger==0.1',
 'pyramid==1.10.4',
 'pytest-cov==2.2.1',
 'pytest-exact-fixtures==0.1',
 'pytest-instafail==0.3.0',
 'pytest-mock==0.11.0',
 'pytest-runner==4.2',
 'pytest-timeout==1.0.0',
 'pytest==2.9.1',
 'python-dateutil==2.5.3',
 'python-magic==0.4.11',
 'pytz==2016.4',
 'rdflib-jsonld==0.4.0',
 'rdflib==4.2.2',
 'requests==2.20.0',
 'rfc3987==1.3.6',
 'rsa==3.4.2',
 'rutter==0.2',
 's3transfer==0.1.13',
 'setuptools>=36.6.0,<37.0.0',
 'simplejson==3.8.2',
 'six==1.14.0',
 'strict-rfc3339==0.7',
 'structlog==18.1.0',
 'subprocess-middleware==0.3',
 'transaction==2.4.0',
 'translationstring==1.3',
 'urllib3==1.23',
 'venusian==1.2.0',
 'waitress==1.2.0',
 'xlrd==1.0.0',
 'zope-deprecation==4.4.0',
 'zope-interface==4.6.0',
 'zope-sqlalchemy==1.2']

entry_points = \
{'console_scripts': ['wipe-test-indices = '
                     'snovault.commands.wipe_test_indices:main']}

setup_kwargs = {
    'name': 'dcicsnovault',
    'version': '2.0.0b0',
    'description': 'Storage support for 4DN Data Portals.',
    'long_description': '========\nSnovault\n========\n\n|Build status|_\n\n.. |Build status| image:: https://travis-ci.org/4dn-dcic/snovault.svg?branch=master\n.. _Build status: https://travis-ci.org/4dn-dcic/snovault\n\nOverview\n========\n\nSnovault is a JSON-LD Database Framework that serves as the backend for the 4DN Data portal and CGAP. Check out our full documentation `here\n<https://snovault.readthedocs.io/en/latest/>`_.\n\n\nInstallation\n============\n\nCurrently these are for Mac OSX using homebrew. If using linux, install dependencies with a different package manager.\n\nStep 0: Install Xcode\n---------------------\n\nInstall Xcode (from App Store) and homebrew: http://brew.sh\n\nStep 1: Verify Homebrew Itself\n------------------------------\n\nVerify that homebrew is working properly::\n\n    $ brew doctor\n\nStep 2: Install Homebrewed Dependencies\n---------------------------------------\n\nInstall or update dependencies::\n\n    $ brew install libevent libmagic libxml2 libxslt openssl postgresql graphviz python3\n    $ brew install freetype libjpeg libtiff littlecms webp  # Required by Pillow\n    $ brew cask install adoptopenjdk8\n    $ brew install elasticsearch@5.6\n\nNOTES:\n\n* If installation of adtopopenjdk8 fails due to an ambiguity, it should work to do this instead::\n\n    $ brew cask install homebrew/cask-versions/adoptopenjdk8\n\n* If you try to invoke elasticsearch and it is not found,\n  you may need to link the brew-installed elasticsearch::\n\n    $ brew link --force elasticsearch@5.6\n\n* If you need to update dependencies::\n\n    $ brew update\n    $ rm -rf encoded/eggs\n\n* If you need to upgrade brew-installed packages that don\'t have pinned versions,\n  you can use the following. However, take care because there is no command to directly\n  undo this effect::\n\n    $ brew update\n    $ brew upgrade\n    $ rm -rf encoded/eggs\n\nStep 3: Running Buildout\n------------------------\n\nRun buildout::\n\n    $ python3 bootstrap.py --buildout-version 2.9.5 --setuptools-version 36.6.0\n    $ bin/buildout\n\n\nNOTES:\n\n* If you have issues with postgres or the python interface to it (psycogpg2)\n  you probably need to install postgresql via homebrew (as above)\n\n* If you have issues with Pillow you may need to install new xcode command line tools.\n\n  - First update Xcode from AppStore (reboot)::\n\n      $ xcode-select --install\n\n  - If you are running macOS Mojave (though this is fixed in Catalina), you may need to run this command as well::\n\n      $ sudo installer -pkg /Library/Developer/CommandLineTools/Packages/macOS_SDK_headers_for_macOS_10.14.pkg -target /\n\n  - If you have trouble with zlib, especially in Catalina, it is probably because brew installed it\n    in a different location. In that case, you\'ll want to do the following\n    in place of the regular call to buildout::\n\n      $ CFLAGS="-I$(brew --prefix zlib)/include" LDFLAGS="-L$(brew --prefix zlib)/lib" bin/buildout\n\n* If you wish to completely rebuild the application, or have updated dependencies,\n  before you go ahead, you\'ll probably want to do::\n\n    $ make clean\n\n  Then goto Step 3.\n\nRunning tests\n=============\n\nTo run specific tests locally::\n\n    $ bin/test -k test_name\n\nTo run with a debugger::\n\n    $ bin/test --pdb\n\nSpecific tests to run locally for schema changes::\n\n    $ bin/test -k test_load_workbook\n\nRun the Pyramid tests with::\n\n    $ bin/test\n\n',
    'author': '4DN-DCIC Team',
    'author_email': 'support@4dnucleome.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/4dn-dcic/snovault',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.4,<4.0',
}


setup(**setup_kwargs)
